from tornado.websocket import WebSocketClosedError
from jupyter_server.services.kernels.connection.base import (
    BaseKernelWebsocketConnection,
)
from .states import LifecycleStates
from jupyter_server.services.kernels.connection.base import deserialize_msg_from_ws_v1, serialize_msg_to_ws_v1

class NextGenKernelWebsocketConnection(BaseKernelWebsocketConnection):
    """A websocket client that connects to a kernel manager.
    
    NOTE: This connection only works with the (newer) v1 websocket protocol.
    https://jupyter-server.readthedocs.io/en/latest/developers/websocket-protocols.html
    """

    kernel_ws_protocol = "v1.kernel.websocket.jupyter.org"

    async def connect(self):
        """A synchronous method for connecting to the kernel via a kernel session.
        This connection might take a few minutes, so we turn this into an
        asyncio task happening in parallel.
        """
        self.kernel_manager.main_client.add_listener(self.handle_outgoing_message)
        await self.kernel_manager.broadcast_state()
        self.log.info("Kernel websocket is now listening to kernel.")

    def disconnect(self):
        self.kernel_manager.main_client.remove_listener(self.handle_outgoing_message)

    def handle_incoming_message(self, incoming_msg):
        """Handle the incoming WS message"""
        channel_name, msg_list = deserialize_msg_from_ws_v1(incoming_msg)
        if self.kernel_manager.main_client:
            self.kernel_manager.main_client.handle_incoming_message(channel_name, msg_list)

    def handle_outgoing_message(self, channel_name, msg):
        """Handle the ZMQ message."""
        try:
            # Remove signature from message to be compatible with Jupyter Server.
            # See here: https://github.com/jupyter-server/jupyter_server/blob/4ee6e1ddc058f87b2c5699cd6b9caf780a013044/jupyter_server/services/kernels/connection/channels.py#L504
            msg = msg[1:]
            msg = serialize_msg_to_ws_v1(msg, channel_name)
            self.websocket_handler.write_message(msg, binary=True)
        except WebSocketClosedError:
            self.log.warning("A ZMQ message arrived on a closed websocket channel.")
        except Exception as err:
            self.log.error(err)


"""
WIP.

This file just contains interfaces to be filled out later.
"""

from __future__ import annotations
from datetime import timedelta, timezone, datetime
from logging import Logger
from typing import TYPE_CHECKING
import uuid
import asyncio

if TYPE_CHECKING:
    from tornado.websocket import WebSocketHandler


class YjsClient:
    """Data model that represents all data associated
    with a user connecting to a YDoc through JupyterLab."""

    websocket: WebSocketHandler
    """The Tornado WebSocketHandler handling the WS connection to this client."""
    id: str
    """UUIDv4 string that uniquely identifies this client."""
    last_modified: datetime
    """Indicates the last modified time when synced state is modified"""

    _synced: bool

    def __init__(self, websocket):
        self.websocket = websocket
        self.id = str(uuid.uuid4())
        self._synced = False
        self.last_modified = datetime.now(timezone.utc)
        

    @property
    def synced(self) -> bool:
        """
        Indicates whether the initial Client SS1 + Server SS2 handshake has been
        completed.
        """
        return self._synced


    @synced.setter
    def synced(self, v: bool) -> None:
        self._synced = v
        self.last_modified = datetime.now(timezone.utc)

class YjsClientGroup:
    """
    Data model that represents a group of clients connected to a room. Provides
    helpful abstractions used by YRoom.
    
    New clients start as desynced. Consumers should call mark_synced() to mark a
    new client as synced once the SS1 + SS2 handshake is complete.
    
    Automatically removes desynced clients if they do not become synced after
    a certain timeout.
    """
    room_id: str
    """Room Id for associated YRoom"""
    synced: dict[str, YjsClient]
    """A dict of client_id and synced YjsClient mapping"""
    desynced: dict[str, YjsClient]
    """A dict of client_id and desynced YjsClient mapping"""
    log: Logger
    """Log object"""
    _poll_interval_seconds: int
    """The poll time interval used while auto removing desynced clients"""
    desynced_timeout_seconds: int
    """The max time period in seconds that a desynced client does not become synced before get auto removed from desynced dict"""
    _stopped: bool
    """
    Whether the client group has been stopped. If `True`, this client group will
    ignore all future calls to `add()`.
    """
    
    def __init__(self, *, room_id: str, log: Logger, poll_interval_seconds: int = 60, desynced_timeout_seconds: int = 120):
        self.room_id = room_id
        self.synced: dict[str, YjsClient] = {}
        self.desynced: dict[str, YjsClient] = {}
        self.log = log
        # asyncio.create_task(self._clean_desynced())
        self._poll_interval_seconds = poll_interval_seconds
        self.desynced_timeout_seconds = desynced_timeout_seconds
        self._stopped = False
        
    def add(self, websocket: WebSocketHandler) -> str | None:
        """
        Adds a new Websocket as a client to the group.

        Returns a client ID if the client group is active.

        Returns `None` if the client group is stopped, which is set after
        `stop()` is called.
        """
        if self._stopped:
            return

        client = YjsClient(websocket)
        self.desynced[client.id] = client
        return client.id
    
    def mark_synced(self, client_id: str) -> None:
        """Marks a client as synced."""
        if client := self.desynced.pop(client_id, None):
            client.synced = True
            self.synced[client.id] = client
    
    def mark_desynced(self, client_id: str) -> None:
        """Marks a client as desynced."""
        if client := self.synced.pop(client_id, None):
            client.synced = False
            self.desynced[client.id] = client

    def remove(self, client_id: str) -> None:
        """Removes a client from the group."""
        client = self.desynced.pop(client_id, None)
        if not client:
            client = self.synced.pop(client_id, None)
        if not client:
            return
        
        try:
            self.log.debug(f"client {client_id} is closed in remove method.")
            client.websocket.close()
        except Exception as e:
            self.log.exception(f"An exception occurred when remove client '{client_id}' for room '{self.room_id}': {e}")  
    
    def get(self, client_id: str) -> YjsClient:
        """
        Gets a client from its ID.
        """
        if client_id in self.desynced: 
            client = self.desynced[client_id]
        if client_id in self.synced:
            client = self.synced[client_id]
        if client.websocket and client.websocket.ws_connection:
            return client
        error_message = f"The client_id '{client_id}' is not found in client group in room '{self.room_id}'"
        self.log.error(error_message)
        raise Exception(error_message)

    def get_all(self, synced_only: bool = True) -> list[YjsClient]:
        """
        Returns a list of all synced clients.
        Set synced_only=False to also get desynced clients.
        """
        all_clients = [c for c in self.synced.values()]
        if not synced_only:
            all_clients += [c for c in self.desynced.values()]
        
        return all_clients

    @property
    def count(self) -> int:
        """Returns the number of clients synced / syncing to the room."""
        return len(self.synced) + len(self.desynced)
    
    async def _clean_desynced(self) -> None:
        while True: 
            try:
                await asyncio.sleep(self._poll_interval_seconds)
                for (client_id, client) in list(self.desynced.items()): 
                    if client.last_modified <= datetime.now(timezone.utc) - timedelta(seconds=self.desynced_timeout_seconds):
                        self.log.warning(f"Remove client '{client_id}' for room '{self.room_id}' since client does not become synced after {self.desynced_timeout_seconds} seconds.")  
                        self.remove(client_id)
                for (client_id, client) in list(self.synced.items()): 
                    if client.websocket is None or client.websocket.ws_connection is None:
                        self.log.warning(f"Remove client '{client_id}' for room '{self.room_id}' since client websocket is closed")  
                        self.remove(client_id)
            except asyncio.CancelledError:
                break
    
    def close_all(self, close_code: int):
        """
        Closes each Websocket with the given close code and removes all clients
        from this group.
        """
        # Remove all clients from both dictionaries
        client_ids = set(self.desynced.keys()) | set(self.synced.keys())
        clients: list[YjsClient] = []
        for client_id in client_ids:
            client = self.desynced.pop(client_id, None)
            if not client:
                client = self.synced.pop(client_id, None)
            if client:
                clients.append(client)
        assert len(self.desynced) == 0
        assert len(self.synced) == 0

        # Close all Websocket connections
        for client in clients:
            client.websocket.close(code=close_code)

    def stop(self, close_code: int = 1001) -> None:
        """
        Closes all Websocket connections with the given close code, removes all
        clients from this group. Future calls to `add()` are ignored until the
        client group is restarted via `restart()`.
         
        If a close code is not specified, it defaults to 1001 (indicates server
        shutting down).
        """
        # Close all Websockets with given close code
        self.close_all(close_code)

        # Set `_stopped` to `True` to ignore future calls to `add()`
        self._stopped = True
    
    @property
    def stopped(self) -> bool:
        """
        Returns whether the client group is stopped.
        """

        return self._stopped
    
    def restart(self, close_code: int = 1001) -> None:
        """
        Restarts the client group by setting `stopped` to `False`. Future calls
        to `add()` will *not* be ignored after this method is called.

        If the client group is not stopped, `self.stop(close_code)` will be
        called with the given argument. Otherwise, `close_code` will be ignored.
        """
        if not self.stopped:
            self.stop(close_code=close_code)

        self._stopped = False
            


import json
import pytest
from pathlib import Path
from tempfile import TemporaryDirectory
from uuid import uuid4

from ..outputs import OutputProcessor, OutputsManager

class OutputProcessorForTest(OutputProcessor):
    """Test subclass of OutputProcessor that overrides the settings property."""
    _test_settings = {}

    @property
    def settings(self):
        """Override the settings property to return a test dictionary."""
        return self._test_settings

@pytest.fixture
def output_processor():
    """Fixture that returns an instance of TestOutputProcessor."""
    return OutputProcessorForTest()

def create_incoming_message(cell_id):
    msg_id = str(uuid4())
    header = {"msg_id": msg_id, "msg_type": "execute_request"}
    parent_header = {}
    metadata = {"cellId": cell_id}
    msg = [json.dumps(item) for item in [header, parent_header, metadata]]
    return msg_id, msg

def test_instantiation(output_processor):
    """Test instantiation of the output processor."""
    op = output_processor
    assert isinstance(op, OutputProcessor)

# TODO: Implement this
def test_output_task():
    pass



import json
import os
from pathlib import Path, PurePath
import shutil
import uuid

from pycrdt import Map
import nbformat

from traitlets.config import LoggingConfigurable
from traitlets import Dict, Instance, Int, default

from jupyter_core.paths import jupyter_runtime_dir


class OutputsManager(LoggingConfigurable):
    _last_output_index = Dict(default_value={})
    _output_index_by_display_id = Dict(default_value={})
    _display_ids_by_cell_id = Dict(default_value={})
    _stream_count = Dict(default_value={})

    outputs_path = Instance(PurePath, help="The local runtime dir")
    stream_limit = Int(default_value=10, config=True, allow_none=True)

    @default("outputs_path")
    def _default_outputs_path(self):
        return Path(jupyter_runtime_dir()) / "outputs"

    def _ensure_path(self, file_id, cell_id):
        nested_dir = self.outputs_path / file_id / cell_id
        nested_dir.mkdir(parents=True, exist_ok=True)

    def _build_path(self, file_id, cell_id=None, output_index=None):
        path = self.outputs_path / file_id
        if cell_id is not None:
            path = path / cell_id
        if output_index is not None:
            path = path / f"{output_index}.output"
        return path
    
    def _compute_output_index(self, cell_id, display_id=None):
        """
        Computes next output index for a cell.
        
        Args:
            cell_id (str): The cell identifier
            display_id (str, optional): A display identifier. Defaults to None.
        
        Returns:
            int: The output index
        """
        last_index = self._last_output_index.get(cell_id, -1)
        if display_id:
            if cell_id not in self._display_ids_by_cell_id:
                self._display_ids_by_cell_id[cell_id] = set([display_id])
            else:
                self._display_ids_by_cell_id[cell_id].add(display_id)
            index = self._output_index_by_display_id.get(display_id)
            if index is None:
                index = last_index + 1
                self._last_output_index[cell_id] = index
                self._output_index_by_display_id[display_id] = index
        else:
            index = last_index + 1
            self._last_output_index[cell_id] = index
        
        return index

    def get_output_index(self, display_id: str):
        """Returns output index for a cell by display_id"""
        return self._output_index_by_display_id.get(display_id)

    def get_output(self, file_id, cell_id, output_index):
        """Get an output by file_id, cell_id, and output_index."""
        path = self._build_path(file_id, cell_id, output_index)
        if not os.path.isfile(path):
            raise FileNotFoundError(f"The output file doesn't exist: {path}")
        with open(path, "r", encoding="utf-8") as f:
            output = json.loads(f.read())
        return output

    def get_outputs(self, file_id, cell_id):
        """Get all outputs by file_id, cell_id."""
        path = self._build_path(file_id, cell_id)
        if not os.path.isdir(path):
            raise FileNotFoundError(f"The output dir doesn't exist: {path}")

        outputs = []

        output_files = [(f, int(f.stem)) for f in path.glob("*.output")]
        output_files.sort(key=lambda x: x[1])
        output_files = output_files[: self.stream_limit]
        has_more_files = len(output_files) >= self.stream_limit

        outputs = []
        for file_path, _ in output_files:
            with open(file_path, "r", encoding="utf-8") as f:
                output = f.read()
                outputs.append(output)

        if has_more_files:
            url = create_output_url(file_id, cell_id)
            placeholder = create_placeholder_dict("display_data", url, full=True)
            outputs.append(json.dumps(placeholder))

        return outputs

    def get_stream(self, file_id, cell_id):
        "Get the stream output for a cell by file_id and cell_id."
        path = self._build_path(file_id, cell_id) / "stream"
        if not os.path.isfile(path):
            raise FileNotFoundError(f"The output file doesn't exist: {path}")
        with open(path, "r", encoding="utf-8") as f:
            output = f.read()
        return output
    
    def write(self, file_id, cell_id, output, display_id=None, asdict: bool = False) -> Map | dict:
        """Write a new output for file_id and cell_id.

        Returns a placeholder output (pycrdt.Map) or None if no placeholder
        output should be written to the ydoc.
        """
        placeholder = self.write_output(file_id, cell_id, output, display_id, asdict=asdict)
        if output["output_type"] == "stream" and self.stream_limit is not None:
            placeholder = self.write_stream(file_id, cell_id, output, placeholder, asdict=asdict)
        return placeholder

    def write_output(self, file_id, cell_id, output, display_id=None, asdict: bool = False) -> Map | dict:
        self._ensure_path(file_id, cell_id)
        index = self._compute_output_index(cell_id, display_id)
        path = self._build_path(file_id, cell_id, index)
        data = json.dumps(output, ensure_ascii=False)
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)
        url = create_output_url(file_id, cell_id, index)
        self.log.info(f"Wrote output: {url}")
        placeholder = create_placeholder_dict(output["output_type"], url)
        if not asdict:
            placeholder = Map(placeholder)
        return placeholder
    
    def write_stream(self, file_id, cell_id, output, placeholder, asdict : bool = False) -> Map | dict:
        # How many stream outputs have been written for this cell previously
        count = self._stream_count.get(cell_id, 0)

        # Go ahead and write the incoming stream
        self._ensure_path(file_id, cell_id)
        path = self._build_path(file_id, cell_id) / "stream"
        text = output["text"]
        with open(path, "a", encoding="utf-8") as f:
            f.write(text)
        url = create_output_url(file_id, cell_id)
        self.log.info(f"Wrote stream: {url}")
        # Increment the count
        count = count + 1
        self._stream_count[cell_id] = count

        # Now create the placeholder output
        if count < self.stream_limit:
            # Return the original placeholder if we haven't reached the limit
            placeholder = placeholder
        elif count == self.stream_limit:
            # Return a link to the full stream output
            placeholder = create_placeholder_dict("display_data", url, full=True)
            if not asdict:
                placeholder = Map(placeholder)
        elif count > self.stream_limit:
            # Return None to indicate that no placeholder should be written to the ydoc
            placeholder = None
        return placeholder

    def clear(self, file_id, cell_id=None):
        """Clear the state of the manager."""
        if cell_id is None:
            self._stream_count = {}
        else:
            self._stream_count.pop(cell_id, None)
            self._last_output_index.pop(cell_id, None)
            
            display_ids = self._display_ids_by_cell_id.get(cell_id, [])
            for display_id in display_ids:
                self._output_index_by_display_id.pop(display_id, None)

        path = self._build_path(file_id, cell_id)    
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            pass

    def process_loaded_notebook(self, file_id: str, file_data: dict) -> dict:
        """Process a loaded notebook and handle outputs through the outputs manager.

        This method processes a notebook that has been loaded from disk.
        If the notebook metadata has placeholder_outputs set to True, 
        outputs are loaded from disk and set as the cell outputs.
        
        Args:
            file_id (str): The file identifier
            file_data (dict): The file data containing the notebook content
                from calling ContentsManager.get()
            
        Returns:
            dict: The modified file data with processed outputs
        """
        self.log.info(f"Processing loaded notebook: {file_id}")

        # Notebook content is a tree of nbformat.NotebookNode objects,
        # which are a subclass of dict.
        nb = file_data['content']
        
        # Check if the notebook metadata has placeholder_outputs set to True
        if nb.get('metadata', {}).get('placeholder_outputs') is True:
            nb = self._process_loaded_placeholders(file_id=file_id, nb=nb)
        else:
            nb = self._process_loaded_no_placeholders(file_id=file_id, nb=nb)           

        file_data['content'] = nb
        return file_data

    def _process_loaded_placeholders(self, file_id: str, nb: dict) -> dict:
        """Process a notebook with placeholder_outputs metadata set to True.
        
        This method processes notebooks that have been saved with placeholder outputs.
        It attempts to load actual outputs from disk and creates placeholder outputs
        for each code cell. If no outputs exist on disk for a cell, the cell's
        outputs are set to an empty list.
        
        Args:
            file_id (str): The file identifier
            nb (dict): The notebook dictionary
            
        Returns:
            dict: The notebook with placeholder outputs loaded from disk
        """
        for cell in nb.get('cells', []):
            if cell.get('cell_type') == 'code':
                cell_id = cell.get('id', str(uuid.uuid4()))
                try:
                    # Try to get outputs from disk
                    output_strings = self.get_outputs(file_id=file_id, cell_id=cell_id)
                    outputs = []
                    for output_string in output_strings:
                        output_dict = json.loads(output_string)
                        placeholder = create_placeholder_dict(
                            output_dict["output_type"],
                            url=create_output_url(file_id, cell_id)
                        )
                        outputs.append(placeholder)
                    cell['outputs'] = outputs
                except FileNotFoundError:
                    # No outputs on disk for this cell, set empty outputs
                    cell['outputs'] = []
        return nb

    def _process_loaded_no_placeholders(self, file_id: str, nb: dict) -> dict:
        """Process a notebook that doesn't have placeholder_outputs metadata.
        
        This method processes notebooks with actual output data in the cells.
        It saves existing outputs to disk and replaces them with placeholder
        outputs that reference the saved files. Outputs that already have
        a URL in their metadata are left as-is.
        
        Args:
            file_id (str): The file identifier
            nb (dict): The notebook dictionary
            
        Returns:
            dict: The notebook with outputs saved to disk and replaced with placeholders
        """
        for cell in nb.get('cells', []):
            if cell.get('cell_type') != 'code' or 'outputs' not in cell:
                continue

            cell_id = cell.get('id', str(uuid.uuid4()))
            processed_outputs = []
            for output in cell.get('outputs', []):
                display_id = output.get('metadata', {}).get('display_id')
                url = output.get('metadata', {}).get('url')
                if url is None:
                    # Save output to disk and replace with placeholder
                    try:
                        placeholder = self.write(
                            file_id,
                            cell_id,
                            output,
                            display_id,
                            asdict=True,
                        )
                    except Exception as e:
                        self.log.error(f"Error writing output: {e}")
                        # If we can't write the output to disk, keep the original
                        placeholder = output
                else:
                    # In this case, there is a placeholder already so keep it
                    placeholder = output
                
                if placeholder is not None:
                    # A placeholder of None means to not add to the YDoc
                    processed_outputs.append(nbformat.from_dict(placeholder))
            
            # Replace the outputs with processed ones
            cell['outputs'] = processed_outputs
        return nb

    def process_saving_notebook(self, nb: dict) -> dict:
        """Process a notebook before saving to disk.

        This method is called when the yroom_file_api saves notebooks.
        It sets the placeholder_outputs key to True in the notebook metadata
        and clears the outputs array for each cell.
        
        Args:
            nb (dict): The notebook dict
            
        Returns:
            dict: The modified file data with placeholder_outputs set to True
                  and empty outputs arrays
        """        
        # Ensure metadata exists
        if 'metadata' not in nb:
            nb['metadata'] = {}
        
        # Set placeholder_outputs to True
        nb['metadata']['placeholder_outputs'] = True
        
        # Clear outputs for all code cells, as they are saved to disk
        for cell in nb.get('cells', []):
            if cell.get('cell_type') == 'code':
                cell['outputs'] = []
        
        return nb


def create_output_url(file_id: str, cell_id: str, output_index: int = None) -> str:
        """Create the URL for an output or stream.

        Parameters:
        - file_id (str): The ID of the file.
        - cell_id (str): The ID of the cell.
        - output_index (int, optional): The index of the output. If None, returns the stream URL.

        Returns:
        - str: The URL string for the output or stream.
        """
        if output_index is None:
            return f"/api/outputs/{file_id}/{cell_id}/stream"
        else:
            return f"/api/outputs/{file_id}/{cell_id}/{output_index}.output"

def create_placeholder_dict(output_type: str, url: str, full: bool = False) -> dict:
    """Build a placeholder output dict for the given output_type and url.
    
    If full is True and output_type is "display_data", returns a display_data output
    with an HTML link to the full stream output.

    Parameters:
    - output_type (str): The type of the output.
    - url (str): The URL associated with the output.
    - full (bool): Whether to create a full output placeholder with a link.

    Returns:
    - dict: The placeholder output dictionary.

    Raises:
    - ValueError: If the output_type is unknown.
    """
    metadata = dict(url=url)
    if full and output_type == "display_data":
        return {
            "output_type": "display_data",
            "data": {
                "text/html": f'<a href="{url}">Click this link to see the full stream output</a>'
            },
        }
    if output_type == "stream":
        return {"output_type": "stream", "text": "", "metadata": metadata}
    elif output_type == "display_data":
        return {"output_type": "display_data", "metadata": metadata}
    elif output_type == "execute_result":
        return {"output_type": "execute_result", "metadata": metadata}
    elif output_type == "error":
        return {"output_type": "error", "metadata": metadata}
    else:
        raise ValueError(f"Unknown output_type: {output_type}")




"""
A new Kernel client that is aware of ydocuments.
"""
import anyio
import asyncio
import json
import typing as t

from traitlets import Set, Instance, Any, Type, default
from jupyter_client.asynchronous.client import AsyncKernelClient

from .message_cache import KernelMessageCache
from jupyter_server_documents.rooms.yroom import YRoom
from jupyter_server_documents.outputs import OutputProcessor
from jupyter_server.utils import ensure_async

from .kernel_client_abc import AbstractDocumentAwareKernelClient


class DocumentAwareKernelClient(AsyncKernelClient):
    """
    A kernel client that routes messages to registered ydocs.
    """
    # Having this message cache is not ideal.
    # Unfortunately, we don't include the parent channel
    # in the messages that generate IOPub status messages, thus,
    # we can't differential between the control channel vs.
    # shell channel status. This message cache gives us
    # the ability to map status message back to their source.
    message_cache = Instance(
        klass=KernelMessageCache
    )

    @default('message_cache')
    def _default_message_cache(self):
        return KernelMessageCache(parent=self)

    # A set of callables that are called when a kernel
    # message is received.
    _listeners = Set(allow_none=True)

    # A set of YRooms that will intercept output and kernel
    # status messages.
    _yrooms: t.Set[YRoom] = Set(trait=Instance(YRoom), default_value=set())


    output_processor = Instance(
        OutputProcessor,
        allow_none=True
    )

    output_process_class = Type(
        klass=OutputProcessor,
        default_value=OutputProcessor
    ).tag(config=True)

    @default("output_processor")
    def _default_output_processor(self) -> OutputProcessor:
        self.log.info("Creating output processor")
        return self.output_process_class(parent=self, config=self.config)

    async def start_listening(self):
        """Start listening to messages coming from the kernel.
        
        Use anyio to setup a task group for listening.
        """
        # Wrap a taskgroup so that it can be backgrounded.
        async def _listening():
            async with anyio.create_task_group() as tg:
                for channel_name in ["shell", "control", "stdin", "iopub"]:
                    tg.start_soon(
                        self._listen_for_messages, channel_name
                    )

        # Background this task.
        self._listening_task = asyncio.create_task(_listening())

    async def stop_listening(self):
        """Stop listening to the kernel.
        """
        # If the listening task isn't defined yet
        # do nothing.
        if not self._listening_task:
            return

        # Attempt to cancel the task.
        try:
            self._listening_task.cancel()
            # Await cancellation.
            await self._listening_task
        except asyncio.CancelledError:
            self.log.info("Disconnected from client from the kernel.")
        # Log any exceptions that were raised.
        except Exception as err:
            self.log.error(err)

    _listening_task: t.Optional[t.Awaitable] = Any(allow_none=True)

    def handle_incoming_message(self, channel_name: str, msg: list[bytes]):
        """Use the given session to send the message."""
        # Cache the message ID and its socket name so that
        # any response message can be mapped back to the
        # source channel.
        header = self.session.unpack(msg[0])
        msg_id = header["msg_id"]                
        metadata = self.session.unpack(msg[2])
        cell_id = metadata.get("cellId")
        
        # Clear cell outputs if cell is re-executedq
        if cell_id:
            existing = self.message_cache.get(cell_id=cell_id)
            if existing and existing['msg_id'] != msg_id:
                asyncio.create_task(self.output_processor.clear_cell_outputs(cell_id))
        
        self.message_cache.add({
            "msg_id": msg_id,
            "channel": channel_name,
            "cell_id": cell_id
        })
        channel = getattr(self, f"{channel_name}_channel")
        channel.session.send_raw(channel.socket, msg)

    def send_kernel_info(self):
        """Sends a kernel info message on the shell channel. Useful 
        for determining if the kernel is busy or idle.
        """
        msg = self.session.msg("kernel_info_request")
        # Send message, skipping the delimiter and signature
        msg = self.session.serialize(msg)[2:]
        self.handle_incoming_message("shell", msg)

    def add_listener(self, callback: t.Callable[[str, list[bytes]], None]):
        """Add a listener to the ZMQ Interface.

        A listener is a callable function/method that takes
        the deserialized (minus the content) ZMQ message.

        If the listener is already registered, it won't be registered again.
        """
        self._listeners.add(callback)

    def remove_listener(self, callback: t.Callable[[str, list[bytes]], None]):
        """Remove a listener. If the listener
        is not found, this method does nothing.
        """
        self._listeners.discard(callback)

    async def _listen_for_messages(self, channel_name: str):
        """The basic polling loop for listened to kernel messages
        on a ZMQ socket.
        """
        # Wire up the ZMQ sockets
        # Setup up ZMQSocket broadcasting.
        channel = getattr(self, f"{channel_name}_channel")
        while True:
            # Wait for a message
            await channel.socket.poll(timeout=float("inf"))
            raw_msg = await channel.socket.recv_multipart()
            # Drop identities and delimit from the message parts.
            _, fed_msg_list = self.session.feed_identities(raw_msg)
            msg = fed_msg_list
            try:
                await self.handle_outgoing_message(channel_name, msg)
            except Exception as err:
                self.log.error(err)

    async def send_message_to_listeners(self, channel_name: str, msg: list[bytes]):
        """
        Sends message to all registered listeners.
        """
        async with anyio.create_task_group() as tg:
            # Broadcast the message to all listeners.
            for listener in self._listeners:
                async def _wrap_listener(listener_to_wrap, channel_name, msg):
                    """
                    Wrap the listener to ensure its async and 
                    logs (instead of raises) exceptions.
                    """
                    try:
                        await ensure_async(listener_to_wrap(channel_name, msg))
                    except Exception as err:
                        self.log.error(err)

                tg.start_soon(_wrap_listener, listener, channel_name, msg)    

    async def handle_outgoing_message(self, channel_name: str, msg: list[bytes]):
        """This is the main method that consumes every
        message coming back from the kernel. It parses the header
        (not the content, which might be large) and updates
        the last_activity, execution_state, and lifecycsle_state
        when appropriate. Then, it routes the message
        to all listeners.
        """
        if channel_name in ('iopub', 'shell'):
            msg = await self.handle_document_related_message(msg)
            # If msg has been cleared by the handler, escape this method.
            if msg is None:
                return
        
        await self.send_message_to_listeners(channel_name, msg)

    async def handle_document_related_message(self, msg: t.List[bytes]) -> t.Optional[t.List[bytes]]:
        """
        Processes document-related messages received from a Jupyter kernel.
        
        Messages are deserialized and handled based on their type. Supported message types
        include updating language info, kernel status, execution state, execution count,
        and various output types. Some messages may be processed by an output processor
        before deciding whether to forward them.

        Returns the original message if it is not processed further, otherwise None to indicate
        that the message should not be forwarded.
        """
        # Begin to deserialize the message safely within a try-except block
        try:
            dmsg = self.session.deserialize(msg, content=False)
        except Exception as e:
            self.log.error(f"Error deserializing message: {e}")
            raise
        
        parent_msg_id = dmsg["parent_header"]["msg_id"]
        parent_msg_data = self.message_cache.get(parent_msg_id)
        cell_id = parent_msg_data.get('cell_id')

        # Handle different message types using pattern matching
        match dmsg["msg_type"]:
            case "kernel_info_reply":
                # Unpack the content to extract language info
                content = self.session.unpack(dmsg["content"])
                language_info = content["language_info"]
                # Update the language info metadata for each collaborative room
                for yroom in self._yrooms:
                    notebook = await yroom.get_jupyter_ydoc()
                    # The metadata ydoc is not exposed as a
                    # public property.
                    metadata = notebook.ymeta
                    metadata["metadata"]["language_info"] = language_info

            case "status":
                # Unpack cell-specific information and determine execution state
                content = self.session.unpack(dmsg["content"])
                execution_state = content.get("execution_state")
                # Update status across all collaborative rooms
                for yroom in self._yrooms:
                    # If this status came from the shell channel, update
                    # the notebook status.
                    if parent_msg_data["channel"] == "shell":                     
                        awareness = yroom.get_awareness()
                        if awareness is not None:
                            # Update the kernel execution state at the top document level
                            awareness.set_local_state_field("kernel", {"execution_state": execution_state})
                    # Specifically update the running cell's execution state if cell_id is provided
                    if cell_id:
                        notebook = await yroom.get_jupyter_ydoc()
                        _, target_cell = notebook.find_cell(cell_id)
                        if target_cell:
                            # Adjust state naming convention from 'busy' to 'running' as per JupyterLab expectation
                            # https://github.com/jupyterlab/jupyterlab/blob/0ad84d93be9cb1318d749ffda27fbcd013304d50/packages/cells/src/widget.ts#L1670-L1678
                            state = 'running' if execution_state == 'busy' else execution_state
                            target_cell["execution_state"] = state
                            break

            case "execute_input":
                if cell_id:
                    # Extract execution count and update each collaborative room's notebook
                    content = self.session.unpack(dmsg["content"])
                    execution_count = content["execution_count"]
                    for yroom in self._yrooms:
                        notebook = await yroom.get_jupyter_ydoc()
                        _, target_cell = notebook.find_cell(cell_id)
                        if target_cell:
                            target_cell["execution_count"] = execution_count
                            break

            case "stream" | "display_data" | "execute_result" | "error" | "update_display_data" | "clear_output":
                if cell_id: 
                    # Process specific output messages through an optional processor
                    if self.output_processor and cell_id:
                        cell_id = parent_msg_data.get('cell_id')
                        content = self.session.unpack(dmsg["content"])
                        self.output_processor.process_output(dmsg['msg_type'], cell_id, content)
                        
                        # Suppress forwarding of processed messages by returning None
                        return None

        # Default return if message is processed and does not need forwarding
        return msg

    async def add_yroom(self, yroom: YRoom):
        """
        Register a YRoom with this kernel client. YRooms will
        intercept display and kernel status messages.
        """
        self._yrooms.add(yroom)

    async def remove_yroom(self, yroom: YRoom):
        """
        De-register a YRoom from handling kernel client messages.
        """
        self._yrooms.discard(yroom)


AbstractDocumentAwareKernelClient.register(DocumentAwareKernelClient)



import json

from jupyter_server.base.handlers import APIHandler
import tornado

from jupyter_server.auth.decorator import authorized
from tornado import web
from tornado.escape import json_encode

from jupyter_server_fileid.manager import BaseFileIdManager

# TODO: This handler belongs in Jupyter Server FileID. 
# Putting it here for now so we don't have to wait for upstream releases.
class FileIDIndexHandler(APIHandler): 
    auth_resource = "contents"

    @property
    def file_id_manager(self) -> BaseFileIdManager:
        return self.settings.get("file_id_manager")
    
    @web.authenticated
    @authorized
    def post(self):
        try:
            path = self.get_argument("path")
            id = self.file_id_manager.index(path)
            self.write(json_encode({"id": id, "path": path}))
        except web.MissingArgumentError:
            raise web.HTTPError(
                400, log_message="'path' parameter was not provided in the request."
            )


class RouteHandler(APIHandler):
    # The following decorator should be present on all verb methods (head, get, post,
    # patch, put, delete, options) to ensure only authorized user can request the
    # Jupyter server
    @tornado.web.authenticated
    def get(self):
        self.finish(json.dumps({
            "data": "This is /jupyter-server-documents/get-example endpoint!"
        }))


import asyncio

from pycrdt import Map

from traitlets import Unicode, Bool, Set
from traitlets.config import LoggingConfigurable
from jupyter_server_documents.kernels.message_cache import KernelMessageCache

class OutputProcessor(LoggingConfigurable):
    
    _file_id = Unicode(default_value=None, allow_none=True)
    _pending_clear_output_cells = Set(default_value=set())

    use_outputs_service = Bool(
        default_value=True,
        help="Should outputs be routed to the outputs service to minimize the in memory ydoc size."
    ).tag(config=True)

    @property
    def settings(self):
        """A shortcut for the Tornado web app settings."""
        #      self.KernelClient.KernelManager.AsyncMultiKernelManager.ServerApp
        return self.parent.parent.parent.parent.web_app.settings

    @property
    def kernel_client(self):
        """A shortcut to the kernel client this output processor is attached to."""
        return self.parent

    @property
    def outputs_manager(self):
        """A shortcut for the OutputsManager instance."""
        return self.settings["outputs_manager"]
    
    @property
    def session_manager(self):
        """A shortcut for the kernel session manager."""
        return self.settings["session_manager"]

    @property
    def file_id_manager(self):
        """A shortcut for the file id manager."""
        return self.settings["file_id_manager"]
    
    @property
    def yroom_manager(self):
        """A shortcut for the jupyter server ydoc manager."""
        return self.settings["yroom_manager"]

    async def get_jupyter_ydoc(self, file_id):
        room_id = f"json:notebook:{file_id}"
        room = self.yroom_manager.get_room(room_id)
        if room is None:
            self.log.error(f"YRoom not found: {room_id}")
            return
        ydoc = await room.get_jupyter_ydoc()
        
        return ydoc

    async def _clear_ydoc_outputs(self, cell_id):
        """Clears the outputs of a cell in ydoc"""
        
        if not self._file_id:
            return
        
        notebook = await self.get_jupyter_ydoc(self._file_id)
        cell_index, target_cell = notebook.find_cell(cell_id)
        if target_cell is not None:
            target_cell["outputs"].clear()
            self.log.info(f"Cleared outputs for {self._file_id=}, {cell_index=}")

    async def clear_cell_outputs(self, cell_id):
        """Clears all outputs for a cell on disk and in ydoc."""

        if self._file_id is not None:
            await self._clear_ydoc_outputs(cell_id)
            self._pending_clear_output_cells.discard(cell_id)
            
            if self.use_outputs_service:
                self.outputs_manager.clear(file_id=self._file_id, cell_id=cell_id)
            

    def process_output(self, msg_type: str, cell_id: str, content: dict):
        """Process outgoing messages from the kernel.
        
        This returns the input dmsg if no the message should be sent to
        clients, or None if it should not be sent.

        The dmsg is a deserialized message generated by calling:

        > self.kernel_client.session.deserialize(dmsg, content=False)

        The content has not been deserialized yet as we need to verify we
        should process it.
        """
        if msg_type == "clear_output":
            asyncio.create_task(self.clear_output_task(cell_id, content))
        else:
            asyncio.create_task(self.output_task(msg_type, cell_id, content))
        
        return None # Don't allow the original message to propagate to the frontend

    async def clear_output_task(self, cell_id, content):
        """A courotine to handle clear_output messages"""

        wait = content.get("wait", False)
        if wait:
            self._pending_clear_output_cells.add(cell_id)
        else:
            await self.clear_cell_outputs(cell_id)

    async def output_task(self, msg_type, cell_id, content):
        """A coroutine to handle output messages."""

        # Check for pending clear_output before processing output
        if cell_id in self._pending_clear_output_cells:
            await self.clear_cell_outputs(cell_id)

        try:
            # TODO: The session manager may have multiple notebooks connected to the kernel
            # but currently get_session only returns the first. We need to fix this and
            # find the notebook with the right cell_id.
            kernel_session = await self.session_manager.get_session(
                kernel_id=self.parent.parent.kernel_id
            )
        except Exception as e:
            self.log.error(
                f"An exception occurred when processing output for cell {cell_id}"
            )
            self.log.exception(e)
            return
        else:
            path = kernel_session["path"]

        file_id = self.file_id_manager.get_id(path)
        if file_id is None:
            self.log.error(f"Could not find file_id for path: {path}")
            return
        self._file_id = file_id

        display_id = content.get("transient", {}).get("display_id")
        # Convert from the message spec to the nbformat output structure
        if self.use_outputs_service: 
            output = self.transform_output(msg_type, content, ydoc=False)
            output = self.outputs_manager.write(
                file_id=file_id,
                cell_id=cell_id,
                output=output,
                display_id=display_id
            )
        else:
            output = self.transform_output(msg_type, content, ydoc=True)

        notebook = await self.get_jupyter_ydoc(file_id)
        if not notebook:
            return
        
        # Write the outputs to the ydoc cell.
        _, target_cell = notebook.find_cell(cell_id)
        if target_cell is not None and output is not None:
            output_index = self.outputs_manager.get_output_index(display_id) if display_id else None
            if output_index is not None:
                target_cell["outputs"][output_index] = output
            else:
                target_cell["outputs"].append(output)
            self.log.info(f"Wrote output to ydoc: {path} {cell_id} {output}")

    
    def transform_output(self, msg_type, content, ydoc=False):
        """Transform output from IOPub messages to the nbformat specification."""
        if ydoc:
            factory = Map
        else:
            factory = lambda x: x
        if msg_type == "stream":
            output = factory({
                "output_type": "stream",
                "text": content["text"],
                "name": content["name"]
            })
        elif msg_type == "display_data" or msg_type == "update_display_data":
            output = factory({
                "output_type": "display_data",
                "data": content["data"],
                "metadata": content["metadata"]
            })
        elif msg_type == "execute_result":
            output = factory({
                "output_type": "execute_result",
                "data": content["data"],
                "metadata": content["metadata"],
                "execution_count": content["execution_count"]
            })
        elif msg_type == "error":
            output = factory({
                "output_type": "error",
                "traceback": content["traceback"],
                "ename": content["ename"],
                "evalue": content["evalue"]
            })
        return output



import pytest
import pytest_asyncio
import shutil
from pathlib import Path
import os
from typing import Awaitable
import pycrdt
from traitlets.config import LoggingConfigurable

from ..rooms import YRoomFileAPI
from jupyter_server.services.contents.filemanager import AsyncFileContentsManager
from jupyter_server_fileid.manager import ArbitraryFileIdManager
from jupyter_ydoc import YUnicode


@pytest.fixture
def jp_contents_manager(tmp_path):
    """
    Returns a configured `ContentsManager` instance whose `root_dir` is set to
    `tmp_path`.

    NOTE: This is a copy of the fixture from jupyter_server, to avoid duplicate
    runs due to parameters in the original fixture.
    """
    return AsyncFileContentsManager(root_dir=str(tmp_path), use_atomic_writing=False)


@pytest.fixture
def fileid_manager(tmp_path):
    """
    Fixture that yields an `ArbitraryFileIdManager` instance whose database file
    is under `{tmp_path}/file_id_manager.db`.
    """
    db_path = os.path.join(tmp_path, "file_id_manager.db")
    return ArbitraryFileIdManager(db_path=db_path)


@pytest.fixture
def mock_plaintext_file(tmp_path):
    """
    Fixture that yields the absolute path to a mock plaintext file under
    `tmp_path`. This mock file has the same content as
    `./mocks/mock_plaintext.txt`.
    """
    # Copy mock file to /tmp
    src_path = Path(__file__).parent / "mocks" / "mock_plaintext.txt"
    target_path = tmp_path / "mock_plaintext.txt"
    shutil.copy(src_path, target_path)

    # Yield the path to the tmp mock plaintext file as a str
    yield str(target_path)

    # Cleanup
    os.remove(target_path)


@pytest_asyncio.fixture(loop_scope="module")
async def plaintext_file_api(
    mock_plaintext_file: str,
    jp_contents_manager: AsyncFileContentsManager,
    fileid_manager: ArbitraryFileIdManager
):
    """
    Returns a `YRoomFileAPI` with a `room_id` corresponding to the
    file created by the `mock_plaintext_file` fixture.
    """
    relpath = os.path.relpath(
        path=mock_plaintext_file,
        start=jp_contents_manager.root_dir
    )
    file_id = fileid_manager.index(relpath)
    room_id = f"text:file:{file_id}"

    class MockYRoom(LoggingConfigurable):
        @property
        def fileid_manager(self):
            return fileid_manager
        
        @property
        def contents_manager(self):
            return jp_contents_manager
        
        @property
        def room_id(self):
            return room_id
        

    yroom_file_api = YRoomFileAPI(
        parent=MockYRoom()
    )
    return yroom_file_api


@pytest.fixture
def empty_yunicode() -> YUnicode:
    """
    Returns an empty `YUnicode` JupyterYDoc.
    """
    ydoc = pycrdt.Doc()
    awareness = pycrdt.Awareness(ydoc=ydoc)
    jupyter_ydoc = YUnicode(ydoc, awareness)
    return jupyter_ydoc


@pytest.mark.asyncio(loop_scope="module")
async def test_load_plaintext_file(
    plaintext_file_api: Awaitable[YRoomFileAPI],
    empty_yunicode: YUnicode,
    mock_plaintext_file: str,
):
    # Load content into JupyterYDoc
    file_api = await plaintext_file_api
    jupyter_ydoc = empty_yunicode
    file_api.load_content_into(jupyter_ydoc)
    await file_api.until_content_loaded
    
    # Assert that the returned JupyterYDoc has the correct content.
    with open(mock_plaintext_file) as f:
        content = f.read()
    assert jupyter_ydoc.source == content
    
    # stop file file api to avoid coroutine warnings
    file_api.stop()




import pytest
from collections import OrderedDict
from jupyter_server_documents.kernels.message_cache import InvalidKeyException, KernelMessageCache, MissingKeyException  # Replace your_module


def create_cache(maxsize=None):
    if maxsize:
        cache = KernelMessageCache(maxsize=maxsize)
    else:
        cache = KernelMessageCache()
    
    # somehow the same cache is shared in the tests
    # clearing the cache so tests pass
    cache._by_msg_id.clear() 
    return cache


def create_message(msg_id, channel, cell_id=None, content="test content"):
    message = {
        "msg_id": msg_id,
        "channel": channel,
        "content": content,
    }
    if cell_id:
        message["cell_id"] = cell_id
    return message


def test_setitem_and_getitem():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    assert cache["msg1"] == message1


def test_setitem_key_mismatch():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    with pytest.raises(InvalidKeyException, match="Key must match `msg_id` in value"):
        cache["wrong_key"] = message1


def test_setitem_missing_msg_id():
    cache = create_cache()
    message1 = {"channel": "shell"}  # Missing msg_id
    with pytest.raises(MissingKeyException, match="`msg_id` missing in message data"):
        cache["key"] = message1


def test_setitem_missing_channel():
    cache = create_cache()
    message1 = {"msg_id": "msg1"}  # Missing channel
    with pytest.raises(MissingKeyException, match="`channel` missing in message data"):
        cache["msg1"] = message1


def test_setitem_with_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    cache["msg1"] = message1
    assert cache._by_cell_id["cell1"] == message1
    assert "msg1" in cache._by_msg_id


def test_setitem_without_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    assert "msg1" in cache._by_msg_id
    assert not cache._by_cell_id


def test_delitem():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    del cache["msg1"]
    assert "msg1" not in cache
    assert "msg1" not in cache._by_msg_id


def test_delitem_with_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    cache["msg1"] = message1
    del cache["msg1"]
    assert "msg1" not in cache
    assert "cell1" not in cache._by_cell_id
    assert "msg1" not in cache._by_msg_id


def test_contains():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    assert "msg1" in cache
    assert "nonexistent_key" not in cache


def test_iter():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    message2 = create_message("msg2", "shell")
    cache["msg1"] = message1
    cache["msg2"] = message2
    keys = list(cache)
    assert "msg1" in keys
    assert "msg2" in keys
    assert len(keys) == 2


def test_len():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    message2 = create_message("msg2", "shell")
    cache["msg1"] = message1
    cache["msg2"] = message2
    assert len(cache) == 2


def test_add():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache.add(message1)
    assert cache["msg1"] == message1
    assert "msg1" in cache._by_msg_id


def test_get_by_msg_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    retrieved_message = cache.get(msg_id="msg1")
    assert retrieved_message == message1
    assert isinstance(cache._by_msg_id, OrderedDict)  # Check it's still OrderedDict


def test_get_by_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    cache["msg1"] = message1
    retrieved_message = cache.get(cell_id="cell1")
    assert retrieved_message == message1
    assert isinstance(cache._by_msg_id, OrderedDict)  # Check it's still OrderedDict


def test_get_not_found():
    cache = create_cache()
    retrieved_message = cache.get(msg_id="nonexistent_key")
    assert retrieved_message is None


def test_remove_by_msg_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    cache.remove(msg_id="msg1")
    assert "msg1" not in cache
    assert "msg1" not in cache._by_msg_id


def test_remove_by_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    cache["msg1"] = message1
    cache.remove(cell_id="cell1")
    assert "msg1" not in cache
    assert "cell1" not in cache._by_cell_id
    assert "msg1" not in cache._by_msg_id


def test_remove_nonexistent():
    cache = create_cache()
    cache.remove(msg_id="nonexistent_key")  # Should not raise an error
    cache.remove(cell_id="nonexistent_cell")  # Should not raise an error


def test_pop_by_msg_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    popped_message = cache.pop(msg_id="msg1")
    assert popped_message == message1
    assert "msg1" not in cache
    assert "msg1" not in cache._by_msg_id


def test_pop_by_cell_id():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    cache["msg1"] = message1
    popped_message = cache.pop(cell_id="cell1")
    assert popped_message == message1
    assert "msg1" not in cache
    assert "cell1" not in cache._by_cell_id
    assert "msg1" not in cache._by_msg_id


def test_pop_nonexistent():
    cache = create_cache()
    with pytest.raises(KeyError):
        cache.pop(msg_id="nonexistent_key")


def test_repr():
    cache = create_cache()
    message1 = create_message("msg1", "shell")
    cache["msg1"] = message1
    representation = repr(cache)
    assert '"msg1":' in representation
    assert '"channel": "shell"' in representation


def test_lru_behavior():
    cache = create_cache()
    cache._by_msg_id = OrderedDict()  # Reset to OrderedDict for LRU test
    cache._by_msg_id["msg1"] = create_message("msg1", "shell")
    cache._by_msg_id["msg2"] = create_message("msg2", "shell")
    cache._by_msg_id["msg3"] = create_message("msg3", "shell")

    # Access "msg1" to make it the most recently used
    cache["msg1"]

    # Check the order after accessing
    expected_order = ["msg2", "msg3", "msg1"]
    assert list(cache._by_msg_id.keys()) == expected_order


def test_maxsize_eviction():
    cache = create_cache(2)
    message1 = create_message("msg1", "shell")
    message2 = create_message("msg2", "shell")
    message3 = create_message("msg3", "shell")

    cache["msg1"] = message1
    cache["msg2"] = message2
    cache["msg3"] = message3  # This should evict "msg1"

    assert "msg1" not in cache
    assert "msg2" in cache
    assert "msg3" in cache
    assert len(cache) == 2


def test_remove_oldest():
    cache = create_cache()
    cache = KernelMessageCache(maxsize=2)
    message1 = create_message("msg1", "shell")
    message2 = create_message("msg2", "shell")
    message3 = create_message("msg3", "shell")
    cache["msg1"] = message1
    cache["msg2"] = message2
    cache["msg3"] = message3 # should trigger remove_oldest

    assert "msg1" not in cache
    assert "msg2" in cache
    assert "msg3" in cache

def test_maxsize_eviction_with_cell_id():
    cache = create_cache(2)
    message1 = create_message("msg1", "shell", "cell1")
    message2 = create_message("msg2", "shell", "cell2")
    message3 = create_message("msg3", "shell", "cell3")

    cache["msg1"] = message1
    cache["msg2"] = message2
    cache["msg3"] = message3  # This should evict "msg1"

    assert "msg1" not in cache
    assert "msg2" in cache
    assert "msg3" in cache
    assert "cell1" not in cache._by_cell_id
    assert "cell2" in cache._by_cell_id
    assert "cell3" in cache._by_cell_id

def test_existing_msg_id_is_removed():
    cache = create_cache()
    message1 = create_message("msg1", "shell", "cell1")
    message2 = create_message("msg2", "shell", "cell1")

    cache["msg1"] = message1
    cache["msg2"] = message2

    assert "msg1" not in cache
    assert message2 == cache["msg2"]


from tempfile import TemporaryDirectory
from pathlib import Path
from uuid import uuid4

import pytest

from ..outputs import OutputsManager


def stream(text: str):
    return {
        "output_type": "stream",
        "name": "stdout",
        "text": text
    }

def display_data_text(text: str):
    return {
        "output_type": "display_data",
        "data": {
            "text/plain": text
        }
}

def test_instantiation():
    op = OutputsManager()
    assert isinstance(op, OutputsManager)

def test_paths():
    """Verify that the paths are working properly."""
    op = OutputsManager()
    file_id = str(uuid4())
    cell_id = str(uuid4())
    with TemporaryDirectory() as td:
        op.outputs_path = Path(td) / "outputs"
        output_index = 0
        assert op._build_path(file_id, cell_id, output_index) == \
            op.outputs_path / file_id / cell_id / f"{output_index}.output"

def test_stream():
    """Test stream outputs."""
    text = "0123456789"
    streams = list([stream(c) for c in text])
    with TemporaryDirectory() as td:
        op = OutputsManager()
        op.outputs_path = Path(td) / "outputs"
        file_id = str(uuid4())
        cell_id = str(uuid4())
        for s in streams:
            placeholder = op.write_output(file_id, cell_id, s)
            op.write_stream(file_id, cell_id, s, placeholder)
        assert op.get_stream(file_id, cell_id) == text

def test_display_data():
    """Test display data."""
    texts = [
        "Hello World!",
        "Hola Mundo!",
        "Bonjour le monde!"
    ]
    outputs = list([display_data_text(t) for t in texts])
    with TemporaryDirectory() as td:
        op = OutputsManager()
        op.outputs_path = Path(td) / "outputs"
        file_id = str(uuid4())
        cell_id = str(uuid4())
        for (i, output) in enumerate(outputs):
            op.write_output(file_id, cell_id, output)
        for (i, output) in enumerate(outputs):
            assert op.get_output(file_id, cell_id, i) == outputs[i]

def test_clear():
    """Test the clearing of outputs for a file_id."""
    output = display_data_text("Hello World!")
    with TemporaryDirectory() as td:
        op = OutputsManager()
        op.outputs_path = Path(td) / "outputs"
        file_id = str(uuid4())
        cell_id = str(uuid4())
        op.write_output(file_id, cell_id, output)
        path = op._build_path(file_id, cell_id, output_index=0)
        assert path.exists()
        op.clear(file_id)
        assert not path.exists()

def file_not_found():
    """Test to ensure FileNotFoundError is raised."""
    with TemporaryDirectory() as td:
        op = OutputsManager()
        op.outputs_path = Path(td) / "outputs"
        with pytest.raises(FileNotFoundError):
            op.get_output('a','b',0)
        with pytest.raises(FileNotFoundError):
            op.get_stream('a','b')       


def test__compute_output_index_basic():
    """
    Test basic output index allocation for a cell without display ID
    """
    op = OutputsManager()
    
    # First output for a cell should be 0
    assert op._compute_output_index('cell1') == 0
    assert op._compute_output_index('cell1') == 1
    assert op._compute_output_index('cell1') == 2

def test__compute_output_index_with_display_id():
    """
    Test output index allocation with display IDs
    """
    op = OutputsManager()
    
    # First output for a cell with display ID
    assert op._compute_output_index('cell1', 'display1') == 0
    
    # Subsequent calls with same display ID should return the same index
    assert op._compute_output_index('cell1', 'display1') == 0
    
    # Different display ID should get a new index
    assert op._compute_output_index('cell1', 'display2') == 1


def test__compute_output_index_multiple_cells():
    """
    Test output index allocation across multiple cells
    """
    op = OutputsManager()
    
    assert op._compute_output_index('cell1') == 0
    assert op._compute_output_index('cell1') == 1
    assert op._compute_output_index('cell2') == 0
    assert op._compute_output_index('cell2') == 1

def test_display_id_index_retrieval():
    """
    Test retrieving output index for a display ID
    """
    op = OutputsManager()
    
    op._compute_output_index('cell1', 'display1')
    
    assert op.get_output_index('display1') == 0
    assert op.get_output_index('non_existent_display') is None

def test_display_ids():
    """
    Test tracking of display IDs for a cell
    """
    op = OutputsManager()
    
    # Allocate multiple display IDs for a cell
    op._compute_output_index('cell1', 'display1')
    op._compute_output_index('cell1', 'display2')
    
    # Verify display IDs are tracked
    assert 'cell1' in op._display_ids_by_cell_id
    assert set(op._display_ids_by_cell_id['cell1']) == {'display1', 'display2'}
    
    # Clear cell indices
    op.clear('file1', 'cell1')
    
    # Verify display IDs are cleared
    assert 'display1' not in op._display_ids_by_cell_id
    assert 'display2' not in op._display_ids_by_cell_id



from __future__ import annotations
from typing import TYPE_CHECKING
import asyncio
from datetime import datetime
from jupyter_ydoc.ybasedoc import YBaseDoc
from jupyter_server.utils import ensure_async
import logging
from tornado.web import HTTPError
from traitlets.config import LoggingConfigurable
from traitlets import Float

if TYPE_CHECKING:
    from typing import Any, Coroutine, Literal
    from .yroom import YRoom
    from jupyter_server_fileid.manager import BaseFileIdManager
    from jupyter_server.services.contents.manager import ContentsManager
    from ..outputs.manager import OutputsManager

class YRoomFileAPI(LoggingConfigurable):
    """
    Provides an API to 1 file from Jupyter Server's ContentsManager for a YRoom.
    This class takes only a single argument: `parent: YRoom`.

    - To load the content, consumers call `load_content_into()` with a
    JupyterYDoc. This also starts the `_watch_file()` loop.

    - Consumers should `await file_api.until_content_loaded` before performing
    any operations on the YDoc.

    - To save a JupyterYDoc to the file, call
    `file_api.schedule_save(jupyter_ydoc)` after calling `load_content_into()`.
    """

    poll_interval = Float(
        default_value=0.5,
        help="Sets how frequently this class saves the YDoc & checks the file "
        "for changes. Defaults to every 0.5 seconds.",
        config=True,
    )

    parent: YRoom
    """
    The parent `YRoom` instance that is using this instance.

    NOTE: This is automatically set by the `LoggingConfigurable` parent class;
    this declaration only hints the type for type checkers.
    """

    log: logging.Logger
    """
    The `logging.Logger` instance used by this class to log.

    NOTE: This is automatically set by the `LoggingConfigurable` parent class;
    this declaration only hints the type for type checkers.
    """

    # See `filemanager.py` in `jupyter_server` for references on supported file
    # formats & file types.
    room_id: str
    file_format: Literal["text", "base64"]
    file_type: Literal["file", "notebook"]
    file_id: str

    _save_scheduled: bool
    _content_loading: bool
    _content_load_event: asyncio.Event

    _last_modified: datetime | None
    """
    The last file modified timestamp known to this instance. If this value
    changes unexpectedly, that indicates an out-of-band change to the file.
    """

    _last_path: str | None
    """
    The last file path known to this instance. If this value changes
    unexpectedly, that indicates an out-of-band move/deletion on the file.
    """

    _watch_file_task: asyncio.Task | None
    """
    The task running the `_watch_file()` loop that processes scheduled saves and
    watches for in-band & out-of-band changes.
    """

    _stopped: bool
    """
    Whether the FileAPI has been stopped, i.e. when the `_watch_file()` task is
    not running.
    """

    _content_lock: asyncio.Lock
    """
    An `asyncio.Lock` that ensures `ContentsManager` calls reading/writing for a
    single file do not overlap. This prevents file corruption scenarios like
    dual-writes or dirty-reads.
    """

    def __init__(self, *args, **kwargs):
        # Forward all arguments to parent class
        super().__init__(*args, **kwargs)

        # Bind instance attributes
        self.file_format, self.file_type, self.file_id = self.room_id.split(":")
        self._save_scheduled = False
        self._last_path = None
        self._last_modified = None
        self._stopped = False

        # Initialize content-related primitives
        self._content_loading = False
        self._content_load_event = asyncio.Event()
        self._content_lock = asyncio.Lock()


    def get_path(self) -> str | None:
        """
        Returns the relative path to the file by querying the FileIdManager. The
        path is relative to the `ServerApp.root_dir` configurable trait.

        Raises a `RuntimeError` if the file ID does not refer to a valid file
        path.
        """
        return self.fileid_manager.get_path(self.file_id)
    
    @property
    def room_id(self) -> str:
        return self.parent.room_id

    @property
    def fileid_manager(self) -> BaseFileIdManager:
        """
        Returns the Jupyter Server's File ID Manager.
        """
        return self.parent.fileid_manager
    
    @property
    def contents_manager(self) -> ContentsManager:
        """
        Stores a reference to the Jupyter Server's ContentsManager.

        NOTE: any calls made on this attribute should acquire & release the
        `_content_lock`. See `_content_lock` for more info.
        """
        return self.parent.contents_manager

    @property
    def outputs_manager(self) -> OutputsManager:
        return self.parent.outputs_manager

    @property
    def content_loaded(self) -> bool:
        """
        Immediately returns whether the YDoc content is loaded.

        To have a coroutine wait until the content is loaded, call `await
        file_api.until_content_loaded` instead.
        """
        return self._content_load_event.is_set()


    @property
    def until_content_loaded(self) -> Coroutine[Any, Any, Literal[True]]:
        """
        Returns an awaitable that resolves when the content is loaded.
        """
        return self._content_load_event.wait()


    def load_content_into(self, jupyter_ydoc: YBaseDoc) -> None:
        """
        Loads the file content into the given JupyterYDoc.
        Consumers should `await file_api.ydoc_content_loaded` before performing
        any operations on the YDoc.

        This method starts the `_watch_file()` task after the content is loaded.
        """
        # If already loaded/loading, return immediately.
        # Otherwise, set loading to `True` and start the loading task.
        if self._content_load_event.is_set() or self._content_loading:
            return
        
        self._content_loading = True
        asyncio.create_task(self._load_content(jupyter_ydoc))

    
    async def _load_content(self, jupyter_ydoc: YBaseDoc) -> None:
        # Get the path specified on the file ID
        path = self.get_path()
        if not path:
            raise RuntimeError(f"Could not find path for room '{self.room_id}'.")
        self._last_path = path

        # Load the content of the file from the path
        self.log.info(f"Loading content for room ID '{self.room_id}', found at path: '{path}'.")
        async with self._content_lock:
            file_data = await ensure_async(self.contents_manager.get(
                path,
                type=self.file_type,
                format=self.file_format
            ))

        if self.file_type == "notebook":
            self.log.info(f"Processing outputs for loaded notebook: '{self.room_id}'.")
            file_data = self.outputs_manager.process_loaded_notebook(file_id=self.file_id, file_data=file_data)

        # Set JupyterYDoc content and set `dirty = False` to hide the "unsaved
        # changes" icon in the UI
        jupyter_ydoc.source = file_data['content']
        jupyter_ydoc.dirty = False

        # Set `_last_modified` timestamp
        self._last_modified = file_data['last_modified']

        # Set loaded event to inform consumers that the YDoc is ready
        # Also set loading to `False` for consistency and log success
        self._content_load_event.set()
        self._content_loading = False
        self.log.info(f"Loaded content for room ID '{self.room_id}'.")

        # Start _watch_file() task
        self._watch_file_task = asyncio.create_task(
            self._watch_file(jupyter_ydoc)
        )


    def schedule_save(self) -> None:
        """
        Schedules a save of the Jupyter YDoc to disk. When called, the Jupyter
        YDoc will be saved on the next tick of the `self._watch_file()`
        background task.
        """
        self._save_scheduled = True
    
    async def _watch_file(self, jupyter_ydoc: YBaseDoc) -> None:
        """
        Defines a background task that processes scheduled saves to the YDoc
        on an interval, checking for in-band & out-of-band changes before doing
        so.

        - The interval duration is set by the `self.poll_interval` configurable
        trait, which defaults to saving & checking every 0.5 seconds.

        - This task is started by a call to `load_ydoc_content()`. Consumers
        must call `self.schedule_save()` for the next tick of this task to save.
        """

        while True:
            try:
                await asyncio.sleep(self.poll_interval)
                await self._check_file()
                if self._save_scheduled:
                    # `asyncio.shield()` prevents the save task from being
                    # cancelled halfway and corrupting the file. We need to
                    # store a reference to the shielded task to prevent it from
                    # being garbage collected (see `asyncio.shield()` docs).
                    save_task = self.save(jupyter_ydoc)
                    await asyncio.shield(save_task)
            except asyncio.CancelledError:
                break
            except Exception:
                self.log.exception(
                    "Exception occurred in `_watch_file() background task "
                    f"for YRoom '{self.room_id}'. Halting for 5 seconds."
                )
                # Wait 5 seconds to reduce error log spam if the exception
                # occurs repeatedly.
                await asyncio.sleep(5)

        self.log.debug(
            "Stopped `self._watch_file()` background task "
            f"for YRoom '{self.room_id}'."
        )

    async def _check_file(self):
        """
        Checks for in-band/out-of-band file operations in the
        `self._watch_file()` background task. This is guaranteed to always run
        before each save in `self._watch_file()` This detects the following
        events and acts in response:

        - In-band move: logs warning (no handling needed)
        - In-band deletion: calls `self.parent.handle_inband_deletion()`
        - Out-of-band move/deletion: calls `self.parent.handle_outofband_move()`
        - Out-of-band change: calls `self.parent.handle_outofband_change()`
        """
        # Ensure that the last known path is defined. This should always be set
        # by `load_ydoc_content()`.
        if not self._last_path:
            raise RuntimeError(f"No last known path for '{self.room_id}'. This should never happen.")

        # Get path. If the path does not match the last known path, the file was
        # moved/deleted in-band via the `ContentsManager`, as it was detected by
        # `jupyter_server_fileid.manager:ArbitraryFileIdManager`.
        # If this happens, run the designated callback and return early.
        path = self.get_path()
        if path != self._last_path:
            if path:
                self.log.warning(
                    f"File was moved to '{path}'. "
                    f"Room ID: '{self.room_id}', "
                    f"Last known path: '{self._last_path}'."
                )
            else:
                self.log.warning(
                    "File was deleted. "
                    f"Room ID: '{self.room_id}', "
                    f"Last known path: '{self._last_path}'."
                )
                self.parent.handle_inband_deletion()
                return

        # Otherwise, set the last known path
        self._last_path = path

        # Build arguments to `CM.get()`
        file_format = self.file_format
        file_type = self.file_type if self.file_type in SAVEABLE_FILE_TYPES else "file"

        # Get the file metadata from the `ContentsManager`.
        # If this raises `HTTPError(404)`, that indicates the file was
        # moved/deleted out-of-band.
        try:
            async with self._content_lock:
                file_data = await ensure_async(self.contents_manager.get(
                    path=path, format=file_format, type=file_type, content=False
                ))
        except HTTPError as e:
            # If not 404, re-raise the exception as it is unknown
            if (e.status_code != 404):
                raise e

            # Otherwise, this indicates the file was moved/deleted out-of-band.
            # Run the designated callback and return early.
            self.log.warning(
                "File was deleted out-of-band. "
                f"Room ID: '{self.room_id}', "
                f"Last known path: '{self._last_path}'."
            )
            self.parent.handle_outofband_move()
            return


        # Finally, if the file was not moved/deleted, check for out-of-band
        # changes to the file content using the metadata.
        # If an out-of-band file change is detected, run the designated callback.
        if self._last_modified != file_data['last_modified']:
            self.log.warning(
                "Out-of-band file change detected. "
                f"Room ID: '{self.room_id}', "
                f"Last detected change: '{self._last_modified}', "
                f"Most recent change: '{file_data['last_modified']}'."
            )
            self.parent.handle_outofband_change()

    
    async def save(self, jupyter_ydoc: YBaseDoc):
        """
        Saves the given JupyterYDoc to disk. This method works even if the
        FileAPI is stopped.

        This method should only be called by consumers if the YDoc needs to be
        saved while the FileAPI is stopped, e.g. when the parent room is
        stopping. In all other cases, consumers should call `schedule_save()`
        instead.
        """
        try:
            # Build arguments to `CM.save()`
            path = self.get_path()
            content = jupyter_ydoc.source
            file_format = self.file_format
            file_type = self.file_type if self.file_type in SAVEABLE_FILE_TYPES else "file"

            # Set `_save_scheduled=False` before the `await` to make sure we
            # save on the next tick when a save is scheduled while `CM.get()` is
            # being awaited.
            self._save_scheduled = False

            if self.file_type == "notebook":
                content = self.outputs_manager.process_saving_notebook(content)

            # Save the YDoc via the ContentsManager
            async with self._content_lock:
                file_data = await ensure_async(self.contents_manager.save(
                    {
                        "format": file_format,
                        "type": file_type,
                        "content": content,
                    },
                    path
                ))

            # Set most recent `last_modified` timestamp
            if file_data['last_modified']:
                self.log.info(f"Reseting last_modified to {file_data['last_modified']}")
                self._last_modified = file_data['last_modified']

            # Set `dirty` to `False` to hide the "unsaved changes" icon in the
            # JupyterLab tab for this YDoc in the frontend.
            jupyter_ydoc.dirty = False
        except Exception as e:
            self.log.error("An exception occurred when saving JupyterYDoc.")
            self.log.exception(e)
    

    def stop(self) -> None:
        """
        Gracefully stops the `YRoomFileAPI`. This immediately halts the
        background task saving the YDoc to the `ContentsManager`.

        To save the YDoc after stopping, call `await
        file_api.save_immediately()` after calling this method.
        """
        if self._watch_file_task:
            self._watch_file_task.cancel()
        self._stopped = True

    @property
    def stopped(self) -> bool:
        """
        Returns whether the FileAPI has been stopped via the `stop()` method.
        """
        return self._stopped

    def restart(self) -> None:
        """
        Restarts the instance by stopping if the room is not stopped, then
        clearing its internal state.

        Consumers should call `load_content_into()` again after this method to
        restart the `_watch_file()` task.
        """
        # Stop if not stopped already
        if not self.stopped:
            self.stop()

        # Reset instance attributes
        self._stopped = False
        self._content_load_event = asyncio.Event()
        self._content_loading = False
        self._save_scheduled = False
        self._last_modified = None
        self._last_path = None
        self.log.info(f"Restarted FileAPI for room '{self.room_id}'.")

    
# see https://github.com/jupyterlab/jupyter-collaboration/blob/main/projects/jupyter-server-ydoc/jupyter_server_ydoc/loaders.py#L146-L149
SAVEABLE_FILE_TYPES = { "directory", "file", "notebook" }



from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager


class NextGenMappingKernelManager(AsyncMappingKernelManager):
    
    def start_watching_activity(self, kernel_id):
        pass
    
    def stop_buffering(self, kernel_id):
        pass